% Aycan Adrian Corum -- aycanc@sabanciuniv.edu

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 0: Introduction
% This code calculates the optimized gamma values for selected beta's in Lasso estimator code.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 1: Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 1a: General Initilization
Var_x = 1;                  % Variation of vector x.
SNR0=10;                    % SNR value, noise level is calculated according to...
sigma0 = sqrt(Var_x/SNR0);  % ... SNR = s*E[xj^2]/sigma0^2 = Var_x/sigma0^2 since vector x is zero mean.
initial_beta = 0.5;         % Starting beta value for testing
end_beta = 3;               % Ending beta value for testing
nb_scalarbetas_opt = 11;    % # of betas between initial_beta and end_beta for optimization.
beta_vec_opt = initial_beta:2.5/(nb_scalarbetas_opt-1):end_beta; % beta vector for optimization.
med_SE_scalar_betavec_opt = zeros(1,nb_scalarbetas_opt);         % This vector will store optimized median SE of asymptotic estimator for each beta.
sigma_effmap_betavec_opt = zeros(1,nb_scalarbetas_opt);          % Stores optimized sigma_effmap values for each beta.
gamma_p_betavec_opt = zeros(1,nb_scalarbetas_opt);               % Stores optimized gamma_p values for each beta.
gamma_betavec_opt = zeros(1,nb_scalarbetas_opt);                 % Stores optimized gamma values for each beta.
%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 1b: Creation of distibution of vector x
% This sub-part is dedicated to the creation of distibution of vector x. Method used here is different and more general (more
% distributions can be applied) than that of used for Lasso distribution so that user may apply any distribution for this
% this asymptotic estimator. The property enabling this generalization is we can use much more data points (create more points from x
% to be estimated). **However, the user should change calculation of variables "prob_axis_for_GaussianSum" and "gamma_betavec_opt(b2)"
% accordingly.
Expectation_samplenb = 500000;                 % # of data points to be created.
ro = 0.1;                                      % Sparsity ratio. (Prob of being non-zero)
x_width = 15;                                  % Starting and ending point of data points x.
x_axis1 = -x_width:2*x_width/(ro*Expectation_samplenb-1):x_width;   % Non-zero part -- data points are with equal distance.
prob_axis1 = normpdf(x_axis1,0,sqrt(Var_x/ro));% Creating probability for non-zero part to have variance Var_x.
prob_axis1 = ro*prob_axis1/sum(prob_axis1);    % Corresponding probabilities for non-zero part.
x_axis2 = zeros(1,(1-ro)*Expectation_samplenb);% Zero part -- data points created at the same point (zero).
prob_axis2 = ones(1,length(x_axis2));          
prob_axis2 = (1-ro)*prob_axis2/sum(prob_axis2);% Corresponding probabilities for zero part.
x_axis = [x_axis1 x_axis2];                    % Combining non-zero and zero parts of distribution of x.
prob_axis = [prob_axis1 prob_axis2];           % Combining corresponding probabilities for non-zero and zero part.
%%%%%%%Comment above and uncomment below to use true prior%%%%%%%%%%%%%%%
% Note that you should change calculation of variables "prob_axis_for_GaussianSum" and "gamma_betavec_opt(b2)" accordingly.
% Expectation_samplenb = 500000;
% c = 3.309722138277721;                         % Coefficient to make variation 1
% x_width = 50;                                  % Starting point is further than that of mixed Ber-Gaus, since true prior dies more slowly.
% x_axis = -x_width:2*x_width/Expectation_samplenb:x_width;
% prob_axis = exp(-c*abs(x_axis).^0.5);          
% prob_axis = prob_axis/sum(prob_axis);
%%%%%%%%%%%%Uncomment until here to use true prior%%%%%%%%%%%%%%%%%%%%%%%
%x_axissq = x_axis.^2;                  % You may uncomment here to check
%real_Var_x = prob_axis*(x_axissq');    % that variance is Var_x.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 2: Optimal Regularization (Calculation of optimized gamma values)
%  As stated in V.D: Optimal Regularization in the paper, for finding the optimal gamma_p that gives the minimum possible MSE, we iteratively minimize 
%  E[s*|x-xhat|^2] wrt gamma_p subject to Pr(|z|>gamma_p/s) < 1/beta. With doing this, we find the minimum value of sigma_effmap by updating 
%  sigma_effmap^2(t) to sigma_effmap^2(t+1) = sigma0^2 + beta*E[s*|x-xhat|^2] after t+1 iterations, where we are certain that it has converged to a value.
%  So, the below loop for each beta follows such a scheme:
%  2.1. Initilization
%  while (sigma_effmap(t) has not converged)
%      2.2. Find minimum possible gamma_p subject to Pr(|z|>gamma_p/s) < 1/beta, called as gamma_p_thres
%      2.3. Find the optimal gamma_p value for given sigma_effmap(t):
%           2.3.1. Start with possible gamma_p values = [10^-3 10^-2.4 ... 10^-0.6 1 10^0.6 ... 10^2.4 10^3] (=gamma_p_innervec_opt) 
%           while(optimal gamma_p has not found with enough certainty)
%               2.3.2. Calculate E[s*|x-xhat|^2] for each gamma_p value in gamma_p_inner_vec_opt
%               2.3.3. Choose gamma_p with minimum E[s*|x-xhat|^2] (or gamma_p_thres if the optimal gamma_p < gamma_p_thres)
%               2.3.4. Scale down the distances in gamma_p_vec_opt by placing newly found optimal gamma_p in the center
%           end
%           2.3.5. Update sigma_effmap by sigma_effmap^2(t) = sigma0^2 + beta*E[s*|x-xhat|^2]             
%  end
%  2.4 Record found minimum MSE and optimal gamma values with their corresponding sigma_effmap and gamma_p values.
 
b2=0; % to keep track of loops
% In the following loop, optimal value of gamma, gamma_p, sigma_effmap and median SE for asymptotic estimator are calculated for each beta:
for beta = beta_vec_opt
    % 2.1 (In-loop initialization):
    b2 = b2+1;
    sigma_effmap = sigma0;           % Initial sigma_effmap.
    previous_sigma_effmap = -10;            % }
    sigma_effmap_vec_opt = zeros(1,1);      % }}  
    sigma_effmap_vec_opt(1) = sigma_effmap; % }}} For showing how sigma_effmap and gamma_p converge/diverge.
    gamma_p_vec_opt = zeros(1,1);           % }}
    gamma_p_vec_opt(1) = 1;                 % }
    t=0;                             % Keeps track of t in sigma_effmap^2(t)
    while abs(sigma_effmap-previous_sigma_effmap) > 0.001                      % while (sigma_effmap(t) has not converged)
        t=t+1;
        previous_sigma_effmap = sigma_effmap;

        % 2.2 (Find minimum possible gamma_p subject to Pr(|z|>gamma_p/s) < 1/beta, called as gamma_p_thres):
        if beta<=1                                                     % If beta <= 1, then Pr(|z|>gamma_p/s) < 1 <= 1/beta is always satisfied...
            gamma_p_thres = 0;                                         % ... since 1 <= 1/beta. Thus gamma_p can be anything.
        else                                                           % else
            gamma_p_axis_for_GaussianSum = 0.001:0.001:10000;          % Axis for possible gamma_p values
            % Find Pr(|z|>gamma_p/s) for each gamma_p in gamma_p_axis_for_GaussianSum:
            prob_axis_for_GaussianSum = 2*((1-ro)*(1-normcdf(gamma_p_axis_for_GaussianSum,0,sigma_effmap))+ro*(1-normcdf(gamma_p_axis_for_GaussianSum,0,sqrt(sigma_effmap^2+1))));
            % And choose smallest gamma_p possible satisfying Pr(|z|>gamma_p/s) < 1/beta:
            prob_axis_for_GaussianSum = prob_axis_for_GaussianSum - 1/beta;
            [dummy min_prob_index] = min(abs(prob_axis_for_GaussianSum));
            if prob_axis_for_GaussianSum(min_prob_index) < 0 
                gamma_p_thres = gamma_p_axis_for_GaussianSum(min_prob_index);
            else
                gamma_p_thres = gamma_p_axis_for_GaussianSum(min_prob_index+1);
            end
        end
        
        % 2.3 (Find the optimal gamma_p value for given sigma_effmap(t)):
        thres_passed=0;                                                % becomes 1 if optimal gamma_p < gamma_p_thres
        E_vec_opt = zeros(1,11);                                       % this vector stores MSE for each gamma_p in gamma_p_innervec_opt
        gamma_p_innervec_opt = zeros(1,11);                            % Initilization of gamma_p_innervec_opt
        scale=10^0.6;                                                  % Starting scaled distance between each gamma_p values in gamma_p_innervec_opt
           % 2.3.1 (Start with possible gamma_p values = [10^-3 10^-2.4 ... 10^-0.6 1 10^0.6 ... 10^2.4 10^3] (=gamma_p_innervec_opt)):
        gamma_p_innervec_opt(1)=1*scale^-5;
        for i=2:11;
            gamma_p_innervec_opt(i)=scale*gamma_p_innervec_opt(i-1);
        end

        max_min_diff = 10;

        nit=0;                                                          % to keep track of inner iterations
        while max_min_diff > 0.01                                       % while(optimal gamma_p has not found with enough certainty)
            nit=nit+1;
            g2 = 0;
            % 2.3.2 (Calculate E[s*|x-xhat|^2] for each gamma_p value in gamma_p_inner_vec_opt):
            for inner_gamma_p = gamma_p_innervec_opt
                g2=g2+1;
                x_E_opt = x_axis';                                      % x_E_opt is the vector x, where x_axis is pre-created distribution.
                v_E_opt = randn(length(x_axis),1);                      % Gaussian noise w/ std. dev. sigma_effmap.
                z_E_opt = x_E_opt + sigma_effmap/1*v_E_opt;             % Measurement vector.
                xhat_scalar_E_opt = zeros(Expectation_samplenb,1);
                for j=1:Expectation_samplenb                            % Recall that the scalar estimator is:
                    if(abs(z_E_opt(j))<=inner_gamma_p)                  % xhat(z) = z-lambda_p = z-gamma_p/s = z-gamma_p, if z>lambda_p
                        xhat_scalar_E_opt(j) = 0;                       %         = z+lambda_p = z+gamma_p/s = z+gamma_p, if z<-lambda_p
                    else                                                %         =                                    0, if |z|<=lambda_p
                        xhat_scalar_E_opt(j) = z_E_opt(j) - sign(z_E_opt(j))*inner_gamma_p;
                    end
                end
                E_vec_opt(g2) = prob_axis*((x_E_opt - xhat_scalar_E_opt).^2);  % Calculate E[s*|x-xhat|^2] over the dist'n, where s=1.
            end

            max_min_diff = max(E_vec_opt) - min(E_vec_opt);             % Update max_min_diff, if it is not small enough, gamma_p_inner_vec_opt...
                                                                        % ...will be further scaled down.

            figure(2);                                                  % Plot the current gamma_p values with their E values.
            subplot(3,1,1);
            plot(gamma_p_innervec_opt,E_vec_opt,'--rs','LineWidth',2,...
               'MarkerEdgeColor','k','MarkerFaceColor','b','MarkerSize',5);
            title(strcat('t= ',num2str(t),', nit= ',num2str(nit),'  (\beta= ',num2str(beta),')'));
            xlabel('\gamma_p');
            ylabel('E = E[s*|x-xhat|^2]');

            % 2.3.3 (Choose gamma_p with minimum E[s*|x-xhat|^2] (or gamma_p_thres if the optimal gamma_p < gamma_p_thres))
            [min_E opt_gamma_p_index] = min(E_vec_opt);                 % Notice that the corresponding E[s*|x-xhat|^2] is stored in min_E
            if gamma_p_innervec_opt(opt_gamma_p_index) > gamma_p_thres  % Choose gamma_p with minimum E[s*|x-xhat|^2]...
                opt_gamma_p = gamma_p_innervec_opt(opt_gamma_p_index);  % ... if gama_p_thres allows.
                thres_passed = 0;                                       % Correct thres_passed if it is accidentally hit in the previous loop.
            else                                                        % else
                opt_gamma_p = gamma_p_thres;                            % we have to choose optimal gamma_p as gamma_p_thres, and...
                if (thres_passed==0)                                    % ... if we are hitting the threshold for the first time...
                    thres_passed = 1;                                   % ... it may be accidental, so give one more chance...
                else                                                    % ... but if the threshold is hit before, then we will not further scale...
                    max_min_diff = 0.001;                               % ... down gamma_p_innervec_opt, since optimal gamma_p is certainly gamma_p_thres
                end                                                     % ... (Making max_min_diff = 0.001 will prevent further iterating)
            end
            % 2.3.4 (Scale down the distances in gamma_p_vec_opt by placing newly found optimal gamma_p in the center)
            scale = scale^0.4;
            gamma_p_innervec_opt(1) = opt_gamma_p*scale^-5;
            for i=2:11;
                gamma_p_innervec_opt(i)=scale*gamma_p_innervec_opt(i-1);
            end

        end

        % If optimal gamma_p is chosen among gamma_p_innervec_opt, then the corresponding E = E[s*|x-xhat|^2] is already calculated and stored in min_E,
        % however if it is gamma_p_thres, then we have not calculated the corresponding E[s*|x-xhat|^2]. The following if-block does that:
        if opt_gamma_p == gamma_p_thres
            x_E_opt = x_axis';
            v_E_opt = randn(length(x_axis),1);
            z_E_opt = x_E_opt + sigma_effmap/1*v_E_opt;
            xhat_scalar_E_opt = zeros(Expectation_samplenb,1);
            for j=1:Expectation_samplenb
                if(abs(z_E_opt(j))<=opt_gamma_p)
                    xhat_scalar_E_opt(j) = 0;
                else
                    xhat_scalar_E_opt(j) = z_E_opt(j) - sign(z_E_opt(j))*opt_gamma_p;
                end
            end
            min_E = prob_axis*((x_E_opt - xhat_scalar_E_opt).^2);
        end

        % 2.3.5 (Update sigma_effmap by sigma_effmap^2(t) = sigma0^2 + beta*E[s*|x-xhat|^2]):
        sigma_effmap = sqrt(sigma0^2 + beta*min_E);
        % Storing calculated sigma_effmap and gamma_p values along the iterations:
        sigma_effmap_vec_opt(t) = sigma_effmap;
        gamma_p_vec_opt(t) = opt_gamma_p;
        
        % Plotting calculated sigma_effmap and gamma_p values along the iterations:
        figure(2)
        subplot(3,1,2)
        plot(gamma_p_vec_opt,'--rs','LineWidth',2,...
                   'MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',4);
        title(strcat('\beta= ',num2str(beta)));
        xlabel('Iteration # (t)');
        ylabel('\gamma_p');
        hold on
        subplot(3,1,3)
        plot(sigma_effmap_vec_opt,'--rs','LineWidth',2,...
                   'MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',4);
        title(strcat('\beta= ',num2str(beta)));
        xlabel('Iteration # (t)');
        ylabel('\sigma_e_f_f_,_m_a_p');
        hold on
    end
    
    % 2.4 (Record found minimum MSE and optimal gamma values with their corresponding sigma_effmap and gamma_p values)
    med_SE_scalar_betavec_opt(b2) = 10*log10(min_E/Var_x); % Record found minimum MSE
    gamma_p_betavec_opt(b2) = opt_gamma_p;                 % Record corresponding gamma_p
    sigma_effmap_betavec_opt(b2) = sigma_effmap;           % Record corresponding sigma_effmap
    % Record most importantly the found optimal gamma_p:
    gamma_betavec_opt(b2)=opt_gamma_p*(1-beta*(2*((1-ro)*(1-normcdf(opt_gamma_p,0,sigma_effmap))+ro*(1-normcdf(opt_gamma_p,0,sqrt(sigma_effmap^2+1))))));
end
optimizedgammaforLASSO_vec=gamma_betavec_opt;   % Record the optimal gamma values also as optimizedgamma_vec, the vector we use in Lasso m-file
% You can also plot median SE for the asymptotic estimator:
figure(1)
plot(beta_vec_opt,med_SE_scalar_betavec_opt);
axis([initial_beta end_beta+0.05 -18 0]); % Same boundaries as in the paper.
xlabel('Measurement ratio \beta = n/m');
ylabel('Median squared error');
hold on