% Aycan Adrian Corum -- aycanc@sabanciuniv.edu

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 0: Introduction
% This code simulates Bernoulli-Gaussian Mixture Distibution for f(x) = |x|^2. Note that the resulting estimator is Linear estimator.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 1: Initialization
% Note that this initialization is for both (Linear and asymptotic) estimators, thus more general. This is the reason for 
% it to be a separate part.
trials = 1000;              % # of Monte Carlo trials.
n = 100;                    % # of data points (dimension of vector x).
S = eye(n);                 % For this case, thse scale factor matrix is identity. 
Var_x = 1;                  % Variation of vector x.
SNR0=10;                    % SNR value, noise level is calculated according to...
sigma0 = sqrt(Var_x/SNR0);  % ... SNR = s*E[xj^2]/sigma0^2 = Var_x/sigma0^2 since vector x is zero mean.
gamma = sigma0^2;           % gamma for Linear estimator, which defines importance of sparsity...
                            % ... Note that gamma = sigma0^2 is the optimized gamma for lowest MSE.
initial_beta = 0.5;         % Starting beta value for testing
end_beta = 3;               % Ending beta value for testing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 2: Linear Estimator
nb_Linbetas = 11;                          % # of betas between initial_beta and end_beta for Linear estimator.
beta_vec = initial_beta:2.5/(nb_Linbetas-1):end_beta;  % beta vector for Linear estimator.
%beta_vec = 0.5; % Uncomment here and change beta value to calculate median SE of Linear for single beta value.
med_SE_Lin_betavec = zeros(1,nb_Linbetas); % This vector will store median SE of Linear estimator for each beta.
ro = 0.1;                                  % Sparsity ratio. (Prob of being non-zero)
r_thres = ceil(ro*1000);
b2=0; % to keep track of loops
% In the following loop, median SE of Linear estimator is calculated for each beta:
for beta = beta_vec
    b2 = b2+1
    SE_Lin_vec = zeros(1,trials); % SE vector to store SE values of each Monte Carlo trial.
    m = ceil(n/beta);               % m as in the paper, # of measurements.
    
    %The following loop executes Monte Carlo measurements #trials times:
    for i = 1:trials
        A = sqrt(1/m)*randn(m,n);   % A as in the paper, the measurement matrix...
                                    % ... The components of A are i.i.d zero mean with variance 1/m.
        x = zeros(n,1);             % Initialization of vector x.
        count=0;                    % Keeps track of number of nonzero elements of x.
        % The following loop creates vector x from Bernoulli-Gaussian mixture distribution:
        for j=1:n;
           r = randi(1000,1);       % A random integer between 1 and 1000 is selected and if it is smaller than ro*1000,...
           if(r <= r_thres)         % ... then that entry of vector x is assigned from a normal distribution...
              x(j)= randn;          % ... So, with "ro" probability x(j) is a standart normal r.v. and and...
              count=count+1;        % ... with (1-ro) probability it is zero.
           end
        end
        x = sqrt(n/count*Var_x)*x;  % Scaling the vector so that overall x has variance 1.

        w = sigma0*randn(m,1);      % The noise w as in the paper, it is Gaussian with variance sigma0.

        y = A*sqrt(S)*x + w;        % Vector of observations as in the paper.

        % Linear estimator:
        xhat_Lin = sqrt(S)*A'/(A*S*A'+gamma*eye(m))*y;
        SE = 20*log10(norm(xhat_Lin-x)/norm(x)); % SE for this Monte Carlo trial.
        SE_Lin_vec(i)=SE;           % Recording this SE value.
    end
    %Now all Monte Carlo trials are done for the current beta value. 
    med_SE_Lin = median(SE_Lin_vec);     % Median SE value among all Monte Carlo trials.
    med_SE_Lin_betavec(b2) = med_SE_Lin; % Store this median SE to corresponding vector.
end
% Now median SE values are calculated for all beta values.
% Now we will plot this:
figure(1)
plot(beta_vec,med_SE_Lin_betavec,'--rs','LineWidth',2,...
            'MarkerEdgeColor','k','MarkerFaceColor','b','MarkerSize',5);
axis([initial_beta end_beta+0.05 -18 0]); % Same boundaries as in the paper.
xlabel('Measurement ratio \beta = n/m');
ylabel('Median squared error');
hold on

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 3: Asymptotic estimator
%%%%%%%%%%%%%%%%%%%%%%%%%
% In this part, we find the corresponding asymptotic estimator for the above Linear estimator, with the fixed
% point equations. These fixed point equations are:
%     Eq1. sigma_effmap^2 = sigma0^2 + beta*E[s*|x-xhat|^2]
%     Eq2. gamma_p = gamma + beta*E[s*sigmasq(z,lambda_p)]
% Since our estimator is linear, we can find the expectations above in closed form as in the paper. Therefore, as shown in
% the paper in page 7 between equations (33) and (34), Eq1 and Eq2 reduces to:
%     Eq1. sigma_effmap^2 = sigma0^2 + beta*E[s*(lambda_p^2*Var_x+mu)/(1+lambda_p)^2]
%     Eq2. gamma_p = gamma + beta*E[s*lambda_p/(1+lambda_p)]
% Since s=1 for our case, these equations further boild down to:
%     Eq1. sigma_effmap^2*(1-beta/(1+gamma_p)^2) = sigma0^2 + beta*gamma_p^2*Var_x/(1+gamma_p)^2
%     Eq2. gamma_p^2 - (beta+gamma-1)*gamma_p - gamma = 0
% As it can be seen, we should first solve Eq2 to solve Eq1. So, we solve these equations after some initalization:

% Genereal initialization:
nb_scalarbetas = 41;                            % # of betas between initial_beta and end_beta for asymptotic estimator.
scalarbeta_vec = initial_beta:2.5/(nb_scalarbetas-1):end_beta;  % beta vector for asymptotic estimator.
%scalarbeta_vec = 1.25; % Uncomment here and change beta value to calculate median SE of asymptotic estimator for single beta value.
med_SE_scalar_betavec = zeros(1,nb_scalarbetas);% This vector will store median SE of asymptotic estimator for each beta.
sigma_effmap_betavec = zeros(1,nb_scalarbetas); % Stores sigma_effmap values for each beta.
gamma_p_betavec = zeros(1,nb_scalarbetas);      % Stores gama_p values for each beta.
b2=0; % to keep track of loops
% In the following loop, median SE of asymptotic estimator is calculated for each beta:
for beta = scalarbeta_vec
    b2=b2+1;
    m = ceil(n/beta);
    gamma_p = ((beta+gamma-1) + sqrt((beta+gamma-1)^2+4*gamma))/2;  % Solve Eq2 for gamma_p
    sigma_effmap = sqrt( (sigma0^2+beta*gamma_p^2*Var_x/(1+gamma_p)^2) / (1-beta/(1+gamma_p)^2) ); % Solve Eq1 for sigma_effmap
    % After calculating both sigma_effmap and gamma_p, the last calculation is the median SE. Recall that SE in dB is equal 
    % to 10*log10(|x-xhat|^2/|x|^2), therefore resulting in median SE = 10*log10(E/Var_x), where as in Eq (33) in the paper E is
    % E = E[|x-xhat|^2] = s*(lambda_p^2*Var_x + mu)/(1+lambda_p)^2 = (gamma_p^2*Var_x + sigma_effmap^2)/(1+gamma_p)^2
    E = (gamma_p^2*Var_x + sigma_effmap^2)/(1+gamma_p)^2;
    med_SE_scalar = 10*log10(E/Var_x);
    med_SE_scalar_betavec(b2) = med_SE_scalar;
    sigma_effmap_betavec(b2) = sigma_effmap; 
    gamma_p_betavec(b2) = gamma_p; 
end

%Finally plot median SE for the asymptotic estimator
figure(1)
plot(scalarbeta_vec,med_SE_scalar_betavec);
axis([initial_beta end_beta+0.05 -18 0]); % Same boundaries as in the paper.
xlabel('Measurement ratio \beta = n/m');
ylabel('Median squared error');
hold on