% Aycan Adrian Corum -- aycanc@sabanciuniv.edu

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 0: Introduction
% This code simulates Bernoulli-Gaussian Mixture Distibution for f(x) = |x|^1. Note that the resulting estimator is the Lasso estimator.
% **IMPORTANT**: Please make "Have_you_run_Lasso_Optimal_Regularization" 1 if you have run "Lasso_Optimal Regularization.m", 0 otherwise:
Have_you_run_Zero_Norm_Regularization_Optimal_Regularization = 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 1: Initialization
% Note that this initialization is for both (Lasso and asymptotic) estimators, thus more general. This is the reason for 
% it to be a separate part.
trials = 1000;              % # of Monte Carlo trials.
n = 100;                    % # of data points (dimension of vector x).
S = eye(n);                 % For this case, thse scale factor matrix is identity. 
Var_x = 1;                  % Variation of vector x.
SNR0=10;                    % SNR value, noise level is calculated according to...
sigma0 = sqrt(Var_x/SNR0);  % ... SNR = s*E[xj^2]/sigma0^2 = Var_x/sigma0^2 since vector x is zero mean.
gamma = 1.5;             % gamma for Zero-Norm Regularized estimator, which defines importance of sparsity.
initial_beta = 0.5;         % Starting beta value for testing
end_beta = 3;               % Ending beta value for testing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 2: Asymptotic estimator
%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 2a: This sub-part is dedicated to the creation of distibution of vector x. Method used here is different and more general (more
% distributions can be applied) than that of used for Lasso distribution so that user may apply any distribution for this
% this asymptotic estimator. The property enabling this generalization is we can use much more data points (create more points from x
% to be estimated). **However, the user should change update of variable "gamma_p" accordingly.
Expectation_samplenb = 500000;                 % # of data points to be created.
ro = 0.1;                                      % Sparsity ratio. (Prob of being non-zero)
x_width = 15;                                  % Starting and ending point of data points x.
x_axis1 = -x_width:2*x_width/(ro*Expectation_samplenb-1):x_width;   % Non-zero part -- data points are with equal distance.
prob_axis1 = normpdf(x_axis1,0,sqrt(Var_x/ro));% Creating probability for non-zero part to have variance Var_x.
prob_axis1 = ro*prob_axis1/sum(prob_axis1);    % Corresponding probabilities for non-zero part.
x_axis2 = zeros(1,(1-ro)*Expectation_samplenb);% Zero part -- data points created at the same point (zero).
prob_axis2 = ones(1,length(x_axis2));          
prob_axis2 = (1-ro)*prob_axis2/sum(prob_axis2);% Corresponding probabilities for zero part.
x_axis = [x_axis1 x_axis2];                    % Combining non-zero and zero parts of distribution of x.
prob_axis = [prob_axis1 prob_axis2];           % Combining corresponding probabilities for non-zero and zero part.
%%%%%%%Comment above and uncomment below to use true prior%%%%%%%%%%%%%%%
% Expectation_samplenb = 500000;
% c = 3.309722138277721;                         % Coefficient to make variation 1
% x_width = 50;                                  % Starting point is further than that of mixed Ber-Gaus, since true prior dies more slowly.
% x_axis = -x_width:2*x_width/Expectation_samplenb:x_width;
% prob_axis = exp(-c*abs(x_axis).^0.5);          
% prob_axis = prob_axis/sum(prob_axis);
%%%%%%%%%%%%Uncomment until here to use true prior%%%%%%%%%%%%%%%%%%%%%%%
%x_axissq = x_axis.^2;                  % You may uncomment here to check
%real_Var_x = prob_axis*(x_axissq');    % that variance is Var_x.
%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 2b: In this sub-part, we find the corresponding asymptotic estimator for the above Lasso, with the fixed
% point equations. These fixed point equations are:
%     Eq1. sigma_effmap^2 = sigma0^2 + beta*E[s*|x-xhat|^2]
%     Eq2. gamma_p = gamma + beta*E[s*sigmasq(z,lambda_p)]
% However, since our estimator is too complicated to find the expectation in Eq1 above in closed form, we use the already created 
% (in part 3a) distributions to find it statistically. In spite of this, we can find the expectation in Eq2 in semi-closed form,
% meaning we cannot find it without iterative method but we do not need to apply statistical methods neither. Therefore the below 
% loop for each beta follows such a scheme:
% 2b1. Initalization
% while (fixed point equations have not converged)
%     2b2. Update Eq2 (As many iterations as needed to make gamma_p converge to a value for the given sigma_effmap/loop)
%     2b3. Update Eq1 (one iteration/loop)
% end
% The reason we apply different iteration numbers for Eq1 and Eq2 is, if we try to apply one iteration/loop to both of them, then
% each may cause the other to diverge. Therefore, we must make one of them settle before updating the other. In this configuration,
% we first settle gamma_p and then update sigma_effmap
% However, we need a general initialization before the initialization in 3b1.

% General initialization:
nb_scalarbetas = 11;                            % # of betas between initial_beta and end_beta for asymptotic estimator.
scalarbeta_vec = initial_beta:2.5/(nb_scalarbetas-1):end_beta;  % beta vector for asymptotic estimator.
%scalarbeta_vec = 1.25; % Uncomment here and change beta value to calculate median SE of asymptotic estimator for single beta value.
med_SE_scalar_betavec = zeros(1,nb_scalarbetas);% This vector will store median SE of asymptotic estimator for each beta.
sigma_effmap_betavec = zeros(1,nb_scalarbetas); % Stores sigma_effmap values for each beta.
gamma_p_betavec = zeros(1,nb_scalarbetas);      % Stores gamma_p values for each beta.
b2=0; % to keep track of loops
% In the following loop, median SE of asymptotic estimator is calculated for each beta:
for beta = scalarbeta_vec
    % 2b1 (In-loop initialization):
    b2=b2+1;
    % Optimized gamma value for the current beta is selected. Of course, this vector is calculated with former calculation in " .m" 
    % If all the results are desired for a specific gamma value, you can do so by making Have_you_run_Zero_Norm_Regularization_Optimal_Regularization = 1
    if Have_you_run_Zero_Norm_Regularization_Optimal_Regularization
        gamma = optimizedgammaforZNR_vec(b2);
    end
    m = ceil(n/beta);
    gamma_p = gamma;                            % Initial gamma_p.
    sigma_effmap = sigma0;                      % Initial sigma_effmap.
    previous_gamma_p = -10;         % }
    previous_sigma_effmap = -10;    % }} For showing how sigma_effmap and    
    gamma_p_vec = zeros(1,1);       % }} gamma_p converge/diverge.
    sigma_effmap_vec = zeros(1,1);  % }
    it=1;
    % The following loop is the essential loop of this part: It keep
    % iterating until both sigma_effmap and gamma_p converges.
    while (abs(gamma_p - previous_gamma_p)>0.001 || abs(sigma_effmap - previous_sigma_effmap)>0.001)
       it                                       % Show the iteration#
       temp_gamma_p = gamma_p;                  % For storing gamma_p and sigma_effmap
       temp_sigma_effmap = sigma_effmap;

       % 2b2 (Updating gamma_p):
       % As stated, as many iterations as needed are done to make gamma_p converge to a value for the given sigma_effmap per loop
       inner_gamma_p = -10;
       g2=0;
       while abs(gamma_p-inner_gamma_p) > 0.1 && g2 < 30
           g2=g2+1;
           inner_gamma_p = gamma_p;
           gamma_p = gamma + beta*gamma_p*( 2*(0.9*(1-normcdf(sqrt(2*gamma_p),0,sigma_effmap))+0.1*(1-normcdf(sqrt(2*gamma_p),0,sqrt(sigma_effmap^2+1)))) );
       end
       
       % 2b3 (Updating sigma_effmap):
       x_E1 = x_axis';                    % x_E1 is the vector x, where x_axis is pre-created distribution.
       v_E1 = randn(length(x_axis),1);    % Gaussian noise w/ std. dev. sigma_effmap.
       z_E1 = x_E1 + sigma_effmap/1*v_E1; % Measurement vector.
       % IMPORTANT: You have already realized that I am creating a vector for tha scalar case, but in the
       % below function xhat_scalar_E, all entries of z_E1 are treated separately, therefore this estimator
       % is exactly a scalar estimator.
       xhat_scalar_E1 = zeros(Expectation_samplenb,1);
       for j=1:Expectation_samplenb                                 % Recall that the scalar estimator is:
            if(abs(z_E1(j))<=sqrt(2*gamma_p))                       % xhat(z) = z, if |z| > sqrt(2*lambda_p)=sqrt(2*gamma_p/s)=sqrt(2*gamma_p)
                xhat_scalar_E1(j) = 0;                              %         = 0, if |z| < sqrt(2*lambda_p)=sqrt(2*gamma_p)
            else                                                  
                xhat_scalar_E1(j) = z_E1(j);
            end
       end
       E1 = prob_axis*((x_E1 - xhat_scalar_E1).^2); % Calculate E[s*|x-xhat|^2] over the dist'n, where s=1.
       sigma_effmap = sqrt(sigma0^2 + beta*E1);     % Update sigma_effmap
       
       % Storing calculated sigma_effmap and gamma_p values:
       sigma_effmap_vec(it) = temp_sigma_effmap;
       gamma_p_vec(it) = temp_gamma_p;
       previous_gamma_p = temp_gamma_p;
       previous_sigma_effmap = temp_sigma_effmap;
       it=it+1;
       % And plotting them, as loops iterate:
       figure(3)
       subplot(2,1,1)
       plot(gamma_p_vec,'--rs','LineWidth',2,'MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',4);
       title(strcat('\beta= ',num2str(beta)));
       ylabel('\gamma_p');
       xlabel('Iteration #');
       hold on
       subplot(2,1,2)
       plot(sigma_effmap_vec,'--rs','LineWidth',2,'MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',4);
       title(strcat('\beta= ',num2str(beta)));
       xlabel('Iteration #');
       ylabel('\sigma_e_f_f_,_m_a_p');
       hold on
    end
   
    % After both sigma_effmap and gamma_p have converged, the last calculation is the median SE. Recall that
    % last calculated E1 is equal to E[|x-xhat|^2]'s, and SE in dB is equal to 10*log10(|x-xhat|^2/|x|^2),
    % therefore resulting SE = 10*log10(E1/Var_x).
    med_SE_scalar = 10*log10(E1/Var_x);
    med_SE_scalar_betavec(b2) = med_SE_scalar;
    sigma_effmap_betavec(b2) = sigma_effmap; 
    gamma_p_betavec(b2) = gamma_p;     
end

%Finally plot median SE for the asymptotic estimator
figure(1)
plot(scalarbeta_vec,med_SE_scalar_betavec);
axis([initial_beta end_beta+0.05 -18 0]); % Same boundaries as in the paper.
xlabel('Measurement ratio \beta = n/m');
ylabel('Median squared error');
hold on